import tkinter
import customtkinter as ctk
import numpy as np

import matplotlib.pyplot as plt

import pandas as pd

from pyrftl.func.wavefront_pair_analysis import wvf_fast_one_pair


class WavefrontConfig(ctk.CTkToplevel):
    # window to display help
    def __init__(self, pair_detail_obj, *args, **kwargs):
        super().__init__()
        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)

        self.title('WavefrontConfig')
        self.geometry("400x500")

        self.pair_detail_obj = pair_detail_obj  # to be able to know current pair

        if "font" in kwargs and kwargs['font'] is not None:
            font = kwargs["font"]
        else:
            font = ctk.CTkFont(size=15)

        # create a frame
        self.main_frame = ctk.CTkFrame(self, fg_color="transparent")
        # self.main_frame.pack(fill=tkinter.BOTH, expand=True)
        self.main_frame.grid(row=0, column=0, sticky="nsew")

        self.main_label = ctk.CTkLabel(self.main_frame, font=font, text="Please select parameters for wavefront.")
        self.main_label.grid(row=0, column=0, padx=5, pady=1, columnspan=3, sticky="w")

        self.label_angle = ctk.CTkLabel(self.main_frame, font=font, text="Field angle [°]")
        self.label_angle.grid(row=2, column=0, padx=5, pady=1, sticky="w")
        cut_off_angle = str(self.pair_detail_obj.param['cut_off_angle_min'])
        self.entry_angle = ctk.CTkEntry(self.main_frame, placeholder_text=cut_off_angle, font=font, width=70)
        self.entry_angle.grid(row=2, column=1, columnspan=1, padx=0, pady=1, sticky="w")

        self.label_grid = ctk.CTkLabel(self.main_frame, font=font, text="Grid side size")
        self.label_grid.grid(row=3, column=0, padx=5, pady=1, sticky="w")
        grid = str(max(self.pair_detail_obj.param['nbRays_first_comparison'],
                       self.pair_detail_obj.param['nbRays_high_comparison']))
        self.entry_grid = ctk.CTkEntry(self.main_frame, placeholder_text=grid, font=font, width=70)
        self.entry_grid.grid(row=3, column=1, columnspan=1, padx=0, pady=1, sticky="w")

        self.checkbox_cyclic = ctk.CTkCheckBox(self.main_frame, text="Display cyclic 1 wave", font=font)
        self.checkbox_cyclic.grid(row=4, column=0, padx=10, pady=1)
        self.checkbox_cyclic.select()  # default ON

        self.checkbox_bestfocus = ctk.CTkCheckBox(self.main_frame, text="Best RMS focus for angle [°]", font=font,
                                                  command=self.checkbox_bestfocus_command)
        self.checkbox_bestfocus.grid(row=5, column=0, padx=0, pady=1)
        self.checkbox_bestfocus.select()  # default ON
        self.entry_bestfocus = ctk.CTkEntry(self.main_frame, placeholder_text="0", font=font, width=70)
        self.entry_bestfocus.grid(row=5, column=1, columnspan=1, padx=0, pady=1, sticky="w")

        self.label_wvf_poly_rms_formula = ctk.CTkLabel(master=self.main_frame, font=font, text="Polychromatic RMS formula")
        self.label_wvf_poly_rms_formula.grid(row=7, column=0, padx=0, pady=1, sticky='e')

        self.optionmenu_wvf_poly_rms_formula = ctk.CTkOptionMenu(self.main_frame, font=font,
                                                                 values=["rms", "worst"])
        self.optionmenu_wvf_poly_rms_formula.grid(row=7, column=1, columnspan=2, padx=0, pady=1, sticky='w')
        self.optionmenu_wvf_poly_rms_formula.set("rms")

        self.label_wvf_focuspoly = ctk.CTkLabel(master=self.main_frame, font=font, text="Best focus polychromatic")
        self.label_wvf_focuspoly.grid(row=6, column=0, padx=0, pady=1, sticky='e')

        self.optionmenu_wvf_focuspoly = ctk.CTkOptionMenu(self.main_frame, font=font,
                                                          values=["reference wavelength", "rms"])
        self.optionmenu_wvf_focuspoly.grid(row=6, column=1, columnspan=2, padx=0, pady=1, sticky='w')
        self.optionmenu_wvf_focuspoly.set("reference wavelength")

        self.label_focus = ctk.CTkLabel(self.main_frame, font=font, text="Focus [mm]")
        self.label_focus.grid(row=8, column=0, padx=5, pady=1, sticky="w")
        self.entry_focus = ctk.CTkEntry(self.main_frame, placeholder_text="0", font=font, width=70)
        self.entry_focus.grid(row=8, column=1, columnspan=1, padx=0, pady=1, sticky="w")
        self.label_focus.grid_remove()
        self.entry_focus.grid_remove()

        self.button_start = ctk.CTkButton(self.main_frame, text="Start", width=20, height=20, font=font,
                                          command=self.start_single)
        self.button_start.grid(row=9, column=0, columnspan=1, padx=0, pady=1, sticky="w")

        # RMS vs fld
        self.checkbox_rmsfld = ctk.CTkCheckBox(self.main_frame, text="RMS for several field angles", font=font,
                                                  command=self.checkbox_rmsfld_command)
        self.checkbox_rmsfld.grid(row=10, column=0, padx=0, pady=1)
        self.checkbox_rmsfld.deselect()

        self.label_anglemin = ctk.CTkLabel(self.main_frame, font=font, text="Angle min [°]")
        self.label_anglemin.grid(row=11, column=0, padx=5, pady=1, sticky="w")
        self.entry_anglemin = ctk.CTkEntry(self.main_frame, placeholder_text="0", font=font, width=70)
        self.entry_anglemin.grid(row=11, column=1, columnspan=1, padx=0, pady=1, sticky="w")

        self.label_anglemax = ctk.CTkLabel(self.main_frame, font=font, text="Angle max [°]")
        self.label_anglemax.grid(row=12, column=0, padx=5, pady=1, sticky="w")
        self.entry_anglemax = ctk.CTkEntry(self.main_frame, placeholder_text=cut_off_angle, font=font, width=70)
        self.entry_anglemax.grid(row=12, column=1, columnspan=1, padx=0, pady=1, sticky="w")

        self.label_anglenb = ctk.CTkLabel(self.main_frame, font=font, text="Number of angle points")
        self.label_anglenb.grid(row=13, column=0, padx=5, pady=1, sticky="w")
        self.entry_anglenb = ctk.CTkEntry(self.main_frame, placeholder_text="5", font=font, width=70)
        self.entry_anglenb.grid(row=13, column=1, columnspan=1, padx=0, pady=1, sticky="w")

        self.label_anglemin.grid_remove()
        self.entry_anglemin.grid_remove()
        self.label_anglemax.grid_remove()
        self.entry_anglemax.grid_remove()
        self.label_anglenb.grid_remove()
        self.entry_anglenb.grid_remove()

        self.button_rmsfld_start = ctk.CTkButton(self.main_frame, text="Start RMSvsField", width=20, height=20,
                                                 font=font, command=self.start_rmsvsfld)
        self.button_rmsfld_start.grid(row=14, column=0, columnspan=1, padx=0, pady=1, sticky="w")
        self.button_rmsfld_start.grid_remove()

        # error message
        self.error_textbox = ctk.CTkTextbox(self.main_frame, wrap="word", font=font, width=400, height=80,
                                            fg_color="transparent", text_color="red")
        self.error_textbox.grid(row=15, column=0, padx=0, pady=5, columnspan=2, sticky="w")
        self.error_textbox.configure(state="disabled")

    def checkbox_bestfocus_command(self):
        # hide or display label and entry
        state = self.checkbox_bestfocus.get()
        if state == 0:  # OFF
            self.entry_bestfocus.grid_remove()  # hide the frame
            self.label_wvf_focuspoly.grid_remove()
            self.optionmenu_wvf_focuspoly.grid_remove()
            self.label_focus.grid()
            self.entry_focus.grid()
        else:  # ON
            self.entry_bestfocus.grid()  # show the frame
            self.label_wvf_focuspoly.grid()
            self.optionmenu_wvf_focuspoly.grid()
            self.label_focus.grid_remove()
            self.entry_focus.grid_remove()

    def checkbox_rmsfld_command(self):
        # hide or display label and entry
        state = self.checkbox_rmsfld.get()
        if state == 0:  # OFF, hide labels
            self.label_anglemin.grid_remove()
            self.entry_anglemin.grid_remove()
            self.label_anglemax.grid_remove()
            self.entry_anglemax.grid_remove()
            self.label_anglenb.grid_remove()
            self.entry_anglenb.grid_remove()
            self.button_rmsfld_start.grid_remove()
        else:  # ON, show labels
            self.label_anglemin.grid()
            self.entry_anglemin.grid()
            self.label_anglemax.grid()
            self.entry_anglemax.grid()
            self.label_anglenb.grid()
            self.entry_anglenb.grid()
            self.button_rmsfld_start.grid()

    def check_entry(self):
        everything_correct = True
        error_txt = ''

        best_focus = self.checkbox_bestfocus.get()
        cyclic = self.checkbox_cyclic.get()

        wvf_angle_entry = self.entry_angle.get()
        if wvf_angle_entry == '':
            wvf_angle = self.pair_detail_obj.param['cut_off_angle_min']
        else :
            try:
                wvf_angle = float(self.entry_angle.get())
            except:
                wvf_angle = None
                everything_correct = False
                error_txt = error_txt + 'Wavefront angle should be a float or nothing. '

        if best_focus:
            focus_dist = None
            focus_angle_entry = self.entry_bestfocus.get()
            if focus_angle_entry == '':
                focus_angle = 0
            else:
                try:
                    focus_angle = float(self.entry_bestfocus.get())
                except:
                    focus_angle = None
                    everything_correct = False
                    error_txt = error_txt + 'Focus angle should be a float or nothing. '
        else :
            focus_angle = None
            focus_dist_entry = self.entry_focus.get()
            if focus_dist_entry == '':
                focus_dist = 0
            else :
                try :
                    focus_dist = float(self.entry_focus.get())
                except:
                    focus_dist = None
                    everything_correct = False
                    error_txt = error_txt + 'Focus distance should be a float or nothing. '

        entry_grid = self.entry_grid.get()
        if entry_grid == '':
            grid = max(self.pair_detail_obj.param['nbRays_first_comparison'],
                       self.pair_detail_obj.param['nbRays_high_comparison'])
        else :
            try:
                grid = int(entry_grid)
            except:
                grid = None
                everything_correct = False
                error_txt = error_txt + 'Grid side size should be an integer or nothing. '

        focus_polychromatic = self.optionmenu_wvf_focuspoly.get()
        match focus_polychromatic :
            case 'reference wavelength' :
                focus_polychromatic = 'ref_wvl_i'
            case "rms":
                focus_polychromatic = 'rmse_rms'
            case _ :
                pass

        polychromatic = self.optionmenu_wvf_poly_rms_formula.get()
        match polychromatic :
            case "rms":
                polychromatic = 'rmse_rms'
            case _ :
                pass

        rmsvsfld = self.checkbox_rmsfld.get()
        if rmsvsfld :
            angle_min = self.entry_anglemin.get()
            if angle_min == '':
                angle_min = 0
            else :
                try:
                    angle_min = float(angle_min)
                except :
                    angle_min = None
                    everything_correct = False
                    error_txt = error_txt + 'Min angle should be float or nothing. '

            angle_max = self.entry_anglemax.get()
            if angle_max == '':
                angle_max = self.pair_detail_obj.param['cut_off_angle_min']
            else :
                try:
                    angle_max = float(angle_max)
                except :
                    angle_max = None
                    everything_correct = False
                    error_txt = error_txt + 'Max angle should be float or nothing. '

            if angle_min > angle_max :
                    angle_min, angle_max = angle_max, angle_min

            angle_nb = self.entry_anglenb.get()
            if angle_nb == '':
                angle_nb = 5
            else :
                try:
                    angle_nb = int(self.entry_anglenb.get())
                    if angle_nb <2 :
                        everything_correct = False
                        error_txt = error_txt + 'The number of angle points should be >= 2. '
                except :
                    angle_nb = None
                    everything_correct = False
                    error_txt = error_txt + 'The number of angle points should be an int or nothing. '

            try :
                angle_list = np.linspace(angle_min, angle_max, angle_nb)
            except Exception as exception:
                angle_nb = angle_list = None
                everything_correct = False
                error_txt = error_txt + 'Impossible to discretize angle interval. Error is printed in Python Console. '
                print('\nImpossible to discretize angle interval. Error is :\n' + str(exception))
        else :
            angle_list = None

        dic_entry = {'wvf_angle':wvf_angle, 'focus_dist':focus_dist, 'focus_angle':focus_angle, 'grid':grid,
                     'best_focus': best_focus, 'cyclic': cyclic, 'polychromatic':polychromatic,
                     'focus_polychromatic':focus_polychromatic, 'angle_list':angle_list}

        # show error text
        self.change_text(error_txt)

        return [everything_correct, error_txt, dic_entry]

    def change_text(self, text):
        # show error text
        self.error_textbox.configure(state="normal")  # configure textbox to be modifiable
        self.error_textbox.delete("0.0", "end")  # delete all text
        self.error_textbox.insert("0.0", text)  # add text
        self.error_textbox.configure(state="disabled")  # configure textbox to be read-only

    def start_single(self):
        everything_correct, error_txt, dic_entry = self.check_entry()

        try :
            if everything_correct :
                # check if need to recompute wavefront or just change the focus
                pair = self.pair_detail_obj.pair

                # compute the wavefront
                param = {'best_focus_fast':dic_entry['best_focus'],
                         'cut_off_angle_min':dic_entry['wvf_angle'], 'save_ram':'no',
                         'focus_polychromatic':dic_entry['focus_polychromatic'],
                         'polychromatic':dic_entry['polychromatic']}

                self.change_text('Processing, please wait...')
                self.update()
                wvf_fast_one_pair(pair, param, nb_rays=dic_entry['grid'], best_focus_angle=dic_entry['focus_angle'],
                                  focus_dist=dic_entry['focus_dist'])
                self.change_text("")

                # print RMS
                print('______________RMS WAVEFRONT ERROR______________')
                print('Pair : ' + pair.short_name)
                print()

                if dic_entry['best_focus']:
                    print('Best focus RMS at angle (°) : ' + str(dic_entry['focus_angle']))
                    print('Focus [mm] : ' + str(pair.rms_detail[0]['foc']))
                    print('Wavefront RMS displayed at angle (°) : ' + str(dic_entry['wvf_angle']))
                    print('Grid side size : ' + str(dic_entry['grid']))
                else :
                    print('Focus [mm] : ' + str(dic_entry['focus_dist']))
                    print('Wavefront RMS displayed at angle (°) : ' + str(dic_entry['wvf_angle']))
                    print('Grid side size : ' + str(dic_entry['grid']))
                print()

                for rms_data in pair.rms_detail :
                    print('Field : ' + str(rms_data['used_fld']))
                    print('Wavelength : ' + str(rms_data['used_wvl']))
                    print('Grid : ' + str(rms_data['grid size']))
                    print('RMS : ' + str(rms_data['RMS']))
                    print('P-V : ' + str(rms_data['P-V']))
                    if rms_data['polychromatic'] is not None :
                        print('RMS polychromatic formula : ' + rms_data['polychromatic'])
                    print()

                print('_______________________________________________')

                # open matplotlib windows with wavefront
                pair.disp_wavefront(cyclic=dic_entry['cyclic'])
        except Exception as exception :
            print('Impossible to display the wavefront, error is :\n' + str(exception))

    def start_rmsvsfld(self):
        everything_correct, error_txt, dic_entry = self.check_entry()

        try :
            if everything_correct and dic_entry['angle_list'] is not None:
                pair = self.pair_detail_obj.pair

                self.change_text('Processing, please wait...')
                self.update()

                angle_list = dic_entry['angle_list']
                rms_poly_list = []

                # compute the first angle
                self.change_text('Processing angle ' + '1/' +
                                 str(len(angle_list)) + ', please wait...')
                self.update()

                param = {'best_focus_fast': dic_entry['best_focus'],
                         'cut_off_angle_min': angle_list[0], 'save_ram': 'no',
                         'focus_polychromatic': dic_entry['focus_polychromatic'],
                         'polychromatic': dic_entry['polychromatic']}
                wvf_fast_one_pair(pair, param, nb_rays=dic_entry['grid'], best_focus_angle=dic_entry['focus_angle'],
                                  focus_dist=dic_entry['focus_dist'])
                wvl_poly = pair.rms_detail[0]['used_wvl']
                rms_poly_list.append(pair.rms_detail[0]['RMS'])


                for i, angle_tmp in enumerate(angle_list[1:], start=1) :
                    self.change_text('Processing angle ' + str(i+1) + '/' +
                                     str(len(angle_list)) + ', please wait...')
                    self.update()

                    # compute the wavefront
                    param = {'best_focus_fast': dic_entry['best_focus'],
                             'cut_off_angle_min': angle_tmp, 'save_ram': 'no',
                             'focus_polychromatic': dic_entry['focus_polychromatic'],
                             'polychromatic': dic_entry['polychromatic']}

                    wvf_fast_one_pair(pair, param, nb_rays=dic_entry['grid'], best_focus_angle=dic_entry['focus_angle'],
                                      focus_dist=dic_entry['focus_dist'])

                    if wvl_poly == pair.rms_detail[0]['used_wvl']:
                        rms_poly_tmp = pair.rms_detail[0]['RMS']
                    else:
                        rms_poly_tmp = np.nan
                    rms_poly_list.append(rms_poly_tmp)

                self.change_text("")

                # print data
                print('______________RMS VS FIELD ANGLE WAVEFRONT ERROR______________')
                print('Pair : ' + pair.short_name)
                print()
                if dic_entry['best_focus']:
                    print('Best focus RMS at angle (°) : ' + str(dic_entry['focus_angle']))
                print('wavelength [nm] : ', wvl_poly)
                print()

                rms_df = pd.DataFrame([[angle_list[i], rms_poly_list[i]] for i in range(len(angle_list))],
                                      columns=['angle', 'rms polychromatic'])
                pd.set_option('display.max_rows', None)  # print all raws of the dataframe
                print(rms_df)
                print()

                print('angle')
                print(angle_list.tolist())
                print('rms polychromatic')
                rms_poly_list_float = [float(rms_i) for rms_i in rms_poly_list]
                print(rms_poly_list_float)



                # matplotlib graph
                plt.figure()
                plt.plot(angle_list, rms_poly_list, color='k', marker='+', linewidth=2, markersize=10,
                         label = 'Polychromatic RMS')
                plt.plot([angle_list[0], angle_list[-1]], [1/14, 1/14], color='k', linestyle='dashed',
                         linewidth=2, label='1/14 waves')
                plt.ylabel('RMS wavefront error [waves]', fontsize=14, fontweight='bold')
                plt.xlabel('Field angle [°]', fontsize=14, fontweight='bold')
                plt.xticks(fontsize=12, fontweight='bold')
                plt.yticks(fontsize=12, fontweight='bold')
                plt.legend()

                plt.show()

        except Exception as exception :
            print('Error while try to compute RMS depending on field angle :\n' + str(exception))


class WavefrontConfigMain :
    # class that can create a WavefrontConfig windows (a windows to select parameters to display a wavefront)
    def __init__(self, pair_detail_obj, font=None):
        self.wavefrontconfig = None
        self.font = font
        self.last_compute = {'pair_short_name':None, 'fld angle':None, 'foc':None, 'grid':None}
        self.pair = None
        self.pair_detail_obj = pair_detail_obj

    def create_wavefrontconfig(self, font=None):
        if (not hasattr(self, 'wavefrontconfig')) or (self.wavefrontconfig is None) or (not self.wavefrontconfig.winfo_exists()):
            if font is None and self.font is not None :
                font = self.font

            self.wavefrontconfig = WavefrontConfig(self.pair_detail_obj, font=font)  # create window if its None or destroyed
            self.wavefrontconfig.after(100, self.wavefrontconfig.focus)  # Workaround for bug where main window takes focus
        else:
            self.wavefrontconfig.focus()



# import from general libraries
import logging
from decimal import Decimal

# import from PyRFTL
from pyrftl.func.various_stuff import is_between

logger = logging.getLogger('PyRFTL')
# log level
LOG_DEBUG = 10
LOG_INFO = 20
LOG_WARNING = 30
LOG_ERROR = 40
LOG_CRITICAL = 50

# which pair to add is the main function to decide which pairs met geometrical distances criteria
# it took two lenses and call which_config_to_add to decide which config of this pair can be used
# this call check_pair_thin_lens to check if the pair met thin lenses distances
# call then check_pair_ray_tracing which check distances with check_pair_dist_ray and then
# check cut-off angle


def check_pair_thin_lens(pair, dist_criteria):
    # this function check if the pair met distances criteria based on thin lens analysis
    # it returns a boolean that said if pair is ok with parameters

    # get a % higher than 0
    if (dist_criteria['tolerance_prcent'] is None) or (
            dist_criteria['tolerance_prcent'] < 0):  # "is None" SHOULD BE before "or ... <0"
        tolerance_prcent = 0
    else:
        tolerance_prcent = dist_criteria['tolerance_prcent']

    # convert the % in multiplying factor
    acc_error_add_mult = 1 + 0.01 * tolerance_prcent
    acc_error_remove_mult = 1 - 0.01 * tolerance_prcent

    # change the min and max distances depending on tolerances
    if dist_criteria['O1O2_min'] is None:
        o1_o2_min = None
    elif dist_criteria['O1O2_min'] < 0:
        o1_o2_min = dist_criteria['O1O2_min'] * acc_error_add_mult
    else:
        o1_o2_min = dist_criteria['O1O2_min'] * acc_error_remove_mult

    if dist_criteria['O1O2_max'] is None:
        o1_o2_max = None
    elif dist_criteria['O1O2_max'] < 0:
        o1_o2_max = dist_criteria['O1O2_max'] * acc_error_remove_mult
    else:
        o1_o2_max = dist_criteria['O1O2_max'] * acc_error_add_mult

    if dist_criteria['ffl_min'] is None:
        ffl_min = None
    elif dist_criteria['ffl_min'] < 0:
        ffl_min = dist_criteria['ffl_min'] * acc_error_add_mult
    else:
        ffl_min = dist_criteria['ffl_min'] * acc_error_remove_mult

    if dist_criteria['ffl_max'] is None:
        ffl_max = None
    elif dist_criteria['ffl_max'] < 0:
        ffl_max = dist_criteria['ffl_max'] * acc_error_remove_mult
    else:
        ffl_max = dist_criteria['ffl_max'] * acc_error_add_mult

    if dist_criteria['bfl_min'] is None:
        bfl_min = None
    elif dist_criteria['bfl_min'] < 0:
        bfl_min = dist_criteria['bfl_min'] * acc_error_add_mult
    else:
        bfl_min = dist_criteria['bfl_min'] * acc_error_remove_mult

    if dist_criteria['bfl_max'] is None:
        bfl_max = None
    elif dist_criteria['bfl_max'] < 0:
        bfl_max = dist_criteria['bfl_max'] * acc_error_remove_mult
    else:
        bfl_max = dist_criteria['bfl_max'] * acc_error_add_mult

    # check criteria

    if ffl_min is None :
        fo1_max = None
    else :
        fo1_max = -ffl_min

    if ffl_max is None:
        fo1_min = None
    else:
        fo1_min = -ffl_max

    good_o1_o2 = is_between(pair.thin_lens_analysis.O1O2, o1_o2_min, o1_o2_max)
    good_fo1 = is_between(pair.thin_lens_analysis.FO1, fo1_min, fo1_max)
    good_o2_fi = is_between(pair.thin_lens_analysis.O2Fi, bfl_min, bfl_max)

    if good_o1_o2 and good_fo1 and good_o2_fi:
        return True
    else:
        return False


def check_pair_dist_ray(pair, dist_criteria):
    # this function check if the pair met distances criteria based on RayOptics analysis
    # it returns a boolean that said if pair is ok with parameters

    # opp_ffl = -pair.ffl
    
    good_ffl = is_between(pair.ffl, dist_criteria['ffl_min'], dist_criteria['ffl_max'])
    good_bfl = is_between(pair.bfl, dist_criteria['bfl_min'], dist_criteria['bfl_max'])
    good_distance_mid = is_between(pair.lenses_gap, dist_criteria['O1O2_min'],
                                   dist_criteria['O1O2_max'])  # distance between two lenses of the pair

    if good_ffl and good_bfl and good_distance_mid:
        return True
    else:
        return False


def check_pair_ray_tracing(pair, param):
    # check with geometrical ray tracing if the pair check distances and cut-off angle criteria

    ####### DISTANCES #######
    # check if distances met criteria
    if not check_pair_dist_ray(pair, param['dist_criteria']):
        why_not_add_pair = ("Pair is rejected by first order ray tracing analysis, it does not meet distances "
                            "criteria. You can change them if needed.")
        return False, why_not_add_pair

    ####### CUT-OFF ANGLE #######
    # check if cut-off angle is wide enough
    # initialisation
    if param['cut_off_angle_test_end'] is None:
        if param['cut_off_angle_min'] < 35 :
            angle_end = 45
        else :
            angle_end = 2 * param['cut_off_angle_min']
    elif param['cut_off_angle_test_end'] < param['cut_off_angle_min']:
        angle_end = param['cut_off_angle_min'] + 2 * param['cut_off_angle_precision']
    else:
        angle_end = param['cut_off_angle_test_end']

    # get cut-off angle of pair
    pair.determine_cutoffangle_dichotomy(param['cut_off_angle_min'], angle_end, param['cut_off_angle_precision'])
    cut_off_angle = pair.cutoffangle['value']

    # check cut-off angle
    if cut_off_angle is None:
        why_not_add_pair = ('Cut-off angle is too small (lower than ' + str(param['cut_off_angle_min']) + ' degrees)')
        return False, why_not_add_pair

    elif cut_off_angle < Decimal(str(param['cut_off_angle_min'])):
        why_not_add_pair = 'Cut-off angle is too small : ' + str(cut_off_angle) + ' degrees'
        return False, why_not_add_pair

    else:  # cut-off angle <= cut_off_angle_min
        return True, ' '



import tkinter
import customtkinter as ctk
import numpy as np
import contextlib, io  # to be able to capture stdout entry, so what is print(...) by a module

from rayoptics.environment import *
import matplotlib.pyplot as plt

from pyrftl.gui.helpbox import HelpWindowMain, create_help_box
from pyrftl.func.save import detail_seq_model_str
from pyrftl.gui.frame import Frame

from opticalglass.schott import SchottGlass
from opticalglass.ohara import OharaGlass
from opticalglass.cdgm import CDGMGlass
from opticalglass.hikari import HikariGlass
from opticalglass.hoya import HoyaGlass
from opticalglass.sumita import SumitaGlass


class RoaFileSelect(ctk.CTkFrame):
    def __init__(self, master, help_window=None, **kwargs):
        # initialize the object
        super().__init__(master, **kwargs)

        # delete previous frame
        if hasattr(master, 'main_frame'):
            master.main_frame.destroy()

        # create a main frame
        master.main_frame = ctk.CTkFrame(master, fg_color="transparent")
        master.main_frame.pack(fill=tkinter.BOTH, expand=True)

        self.main_frame = master.main_frame

        if hasattr(master, 'font'):
            font = master.font
        else:
            font = ctk.CTkFont(size=15, weight="bold")

        if help_window is not None:
            self.help_window = help_window
        else:
            self.help_window = HelpWindowMain(font)

        self.textbox_pair_short_info = ctk.CTkTextbox(self.main_frame, wrap="word", font=font, width=500,
                                                      height=60)
        self.textbox_pair_short_info.grid(row=1, column=0, padx=5, pady=(0,10), columnspan=2, sticky="w")
        self.textbox_pair_short_info.configure(state="disabled")  # configure textbox to be read-only

        self.button_open_file = ctk.CTkButton(self.main_frame, text="Open roa file", command=self.open_roa_file,
                                                font=font)
        self.button_open_file.grid(row=1, column=2, padx=10, pady=(10,30), sticky='n')

        self.textbox_pair_short_info.configure(state="normal")  # configure textbox to be modifiable
        self.textbox_pair_short_info.insert("0.0", 'Please select a .roa file to be displayed.')  # add text
        self.textbox_pair_short_info.configure(state="disabled")  # configure textbox to be read-only

        # frame to show the model
        self.opm_detail_frame = ctk.CTkFrame(self.main_frame)
        self.opm_detail_frame.grid(row=2, column=0, columnspan=3)
        self.opm_detail_frame.grid_remove()  # hide frame
        self.opm_detail_obj = OPMDetail(self.opm_detail_frame, opm=None, help_window=self.help_window)


    def open_roa_file(self):
        # get the filepath of the model file to open
        filepath = tkinter.filedialog.askopenfilename(filetypes=[("RayOptics", ".roa")])
        print(filepath)

        try:
            if filepath == '' :
                # cancel the selection of a file
                pass
            else :
                self.filepath = filepath
                # open the optical model
                opm = open_model(self.filepath)
                print(opm)

                # change the displayed informations
                self.opm_detail_obj.change_opm(opm)

                # show the frame if it was hidden
                self.opm_detail_frame.grid()

                # change information in the short info box
                txt = self.filepath

        except Exception as exception:
            txt = 'Error when try to open ' + filepath + '\nError is : ' + str(exception)

            # hide the optical model frame
            self.opm_detail_frame.grid_remove()
            self.opm_detail_obj.change_opm(None)

        if not filepath == '':
            self.textbox_pair_short_info.configure(state="normal")  # configure textbox to be modifiable
            self.textbox_pair_short_info.delete("0.0", "end")
            self.textbox_pair_short_info.insert("0.0", txt)  # add text
            self.textbox_pair_short_info.configure(state="disabled")  # configure textbox to be read-only


class OPMDetail(Frame):
    # frame with a textbox which show some details of the optical model, and allow to modify it
    def __init__(self, master, opm, new_frame=False, help_window=None, **kwargs):
        # initialize the object
        super().__init__(master, new_frame, help_window, **kwargs)

        self.label_details_text = ctk.CTkLabel(self.frame, font=self.font, text="Optical model details :")
        self.label_details_text.grid(row=0, column=0, padx=5, pady=1, columnspan=2, sticky="w")

        self.textbox_details = ctk.CTkTextbox(self.frame, wrap="word", width=700, height=350,
                                              font=ctk.CTkFont(family="Lucida Console", size=14))  # font with same space for all characters
        self.textbox_details.grid(row=0, column=0, padx=5, pady=(0,10), columnspan=5, sticky="w")
        self.textbox_details.configure(state="disabled")  # configure textbox to be read-only

        self.opm = None
        self.sm = self.osp = self.pm = self.em = self.pt = self.ar = None

        if opm is not None:
            self.change_opm(opm)
        else:
            self.update_textbox()

        self.button_ray = ctk.CTkButton(self.frame, text="Ray diagram", command=self.raydiagram, font=self.font)
        self.button_ray.grid(row=1, column=0, padx=10, pady=(5,10), sticky='n')

        self.button_wvf = ctk.CTkButton(self.frame, text="Wavefront", command=self.wavefront, font=self.font)
        self.button_wvf.grid(row=1, column=1, padx=10, pady=(5,10), sticky='n')

        self.button_wvf = ctk.CTkButton(self.frame, text="Save model .roa", command=self.save_roa, font=self.font)
        self.button_wvf.grid(row=1, column=2, padx=10, pady=(5,10), sticky='n')

        self.modify_label = ctk.CTkLabel(self.frame, text='Modify model :', font=self.font)
        self.modify_label.grid(row=2, column=0, columnspan=2, sticky='w', padx=10, pady=5)

        self.button_wvf = ctk.CTkButton(self.frame, text="Interfaces", command=self.save_roa, font=self.font)
        self.button_wvf.grid(row=3, column=0, padx=10, pady=(0,30), sticky='n')

        self.button_wvf = ctk.CTkButton(self.frame, text="Gaps", command=self.save_roa, font=self.font)
        self.button_wvf.grid(row=3, column=1, padx=10, pady=(0,30), sticky='n')

        self.button_wvf = ctk.CTkButton(self.frame, text="Specs", command=self.save_roa, font=self.font)
        self.button_wvf.grid(row=3, column=2, padx=10, pady=(0,30), sticky='n')

        # initialize glass catalogs
        # if not, medium instances of glass type will have None as catalog attribute
        SchottGlass.initialize_catalog(None)
        SumitaGlass.initialize_catalog(None)
        OharaGlass.initialize_catalog(None)
        HikariGlass.initialize_catalog(None)
        HoyaGlass.initialize_catalog(None)
        CDGMGlass.initialize_catalog(None)


    def update_textbox(self):
        if self.opm is None :
            detail_txt = ('No optical model has been selected, nothing to display !\n'
                          'Please use the "Open roa file" button to select a model.')
        else:
            detail_txt = detail_opm_seq_model_str(self.opm)

        self.textbox_details.configure(state="normal")  # configure textbox to be modifiable
        self.textbox_details.delete("0.0", "end")  # delete all text
        self.textbox_details.insert("0.0", detail_txt)  # add text
        self.textbox_details.configure(state="disabled")  # configure textbox to be read-only

    def change_opm(self, new_opm):
        try:
            if new_opm is not None:
                self.sm = new_opm['seq_model']
                self.osp = new_opm['optical_spec']
                self.pm = new_opm['parax_model']
                self.em = new_opm['ele_model']
                self.pt = new_opm['part_tree']
                self.ar = new_opm['analysis_results']
        except Exception as exception:
            new_opm = None

        if new_opm is None:
            self.sm = self.osp = self.pm = self.em = self.pt = self.ar = None

        self.opm = new_opm

        self.update_textbox()

    def raydiagram(self):
        # display the pair ray tracing diagram
        if self.opm is not None:
            try :
                fig_raydiagram = plt.figure(FigureClass=InteractiveLayout, opt_model=self.opm,
                                            do_draw_rays=True, do_paraxial_layout=True).plot()
                title = 'YZ diagram'
                fig_raydiagram.suptitle(title)
                window_name = 'YZ diagram'
                fig_raydiagram.canvas.manager.set_window_title(window_name)
            except :
                pass

    def wavefront(self):
        plt.figure()


        if self.opm is not None:
            try :
                plt.plot([i for i in range(5)], [i for i in range(5)])
            except:
                pass

    def save_roa(self):
        try:
            filepath = tkinter.filedialog.asksaveasfilename(defaultextension='.roa')
            self.opm.save_model(filepath)
        except:
            pass


def detail_opm_seq_model_str(opm):
    more_info = ''
    more_info = more_info +"List sequential model :\n"

    sm = opm['seq_model']
    osp = opm['optical_spec']
    pm = opm['parax_model']

    # capture the output of listing sequential model (which is output with print, not in str)
    # https://stackoverflow.com/questions/22822267/how-to-capture-print-output-of-another-module
    # https://docs.python.org/3/library/contextlib.html
    f = io.StringIO()
    with contextlib.redirect_stdout(f):
        sm.list_model()
    list_model_str = f.getvalue()
    more_info = more_info + list_model_str + "\n\n"

    # pupil and wavelength etc
    more_info = more_info + 'system unit : ' + opm.system_spec.dimensions + '\n\n'
    more_info = more_info + 'wavelengths (nm) : ' + str(osp.spectral_region.wavelengths) + '\n\n'
    more_info = more_info + 'pupil diameter : ' + osp['pupil'].listobj_str() + '\n\n'
    more_info = more_info + 'fields angles : ' + osp['fov'].listobj_str() + '\n\n'

    # first order data of the model => capture string output
    f = io.StringIO()
    with contextlib.redirect_stdout(f):
        pm.first_order_data()
    first_order_data_str = f.getvalue()
    more_info = more_info + 'first order data :\n' + first_order_data_str + "\n\n"

    # detailed information about surfaces and gap
    more_info = more_info + ("More information about surfaces and gaps (coefficients if some surfaces are asphere, "
                             "refractive index used, and more) :\n\n")

    for i in range(len(sm.ifcs)):
        more_info = (more_info + "INTERFACE " + str(i) + " :\n" +
                     sm.ifcs[i].listobj_str() + "\n")

        if i < len(sm.ifcs) - 1:
            more_info = (more_info + "GAP " + str(i) + " :\n" +
                         sm.gaps[i].listobj_str())

            for wvl in opm['optical_spec'].spectral_region.wavelengths:
                more_info = (more_info + "RI of medium for wvl " + str(wvl) + " nm is : " +
                             str(sm.gaps[i].medium.calc_rindex(wvl)) + "\n")
            more_info = more_info + "\n"
    return more_info




import tkinter
import customtkinter as ctk


class HelpBox(ctk.CTkToplevel):
    # window to display help
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)

        self.title('Help')
        self.geometry("400x300")

        if "text" in kwargs :
            text = kwargs["text"]
        else:
            text = ''

        if "image" in kwargs :
            image = kwargs["image"]
        else:
            image = None

        if "font" in kwargs and kwargs['font'] is not None:
            font = kwargs["font"]
        else:
            font = ctk.CTkFont(size=15)

        # create a frame
        self.main_frame = ctk.CTkFrame(self, fg_color="transparent")
        # self.main_frame.pack(fill=tkinter.BOTH, expand=True)
        self.main_frame.grid(row=0, column=0, sticky="nsew")
        self.main_frame.grid_columnconfigure(0, weight=1)
        self.main_frame.grid_rowconfigure(0, weight=1)

        # create textbox
        self.textbox = ctk.CTkTextbox(self.main_frame, wrap="word", font=font)
        # self.textbox.pack(fill=tkinter.BOTH, expand=True)
        self.textbox.grid(row=0, rowspan=2, column=0, sticky="nsew")

        # add text
        self.textbox.insert("0.0", text)
        self.textbox.configure(state="disabled")  # configure textbox to be read-only

        # create image
        if image is not None:
            self.label_image = ctk.CTkLabel(self.main_frame, text="", image=image)
            self.label_image.grid(row=0, column=1)
            # self.label_image.grid_remove()
        else:
            self.label_image=None

    def text(self, text, image=None, font=None):
        # change the text
        self.textbox.configure(state="normal")  # configure textbox to be modifiable
        self.textbox.delete("0.0", "end")  # delete all text
        self.textbox.insert("0.0", text)  # add text
        self.textbox.configure(state="disabled")  # configure textbox to be read-only

        # change image
        if self.label_image is not None:
            self.label_image.destroy()
            self.label_image = None
        if image is not None:
            self.label_image = ctk.CTkLabel(self.main_frame, text="", image=image)
            self.label_image.grid(row=0, column=1)

        self.focus()  # focus on the windows


def create_help_box(obj, text, image=None, font=None):
    if (not hasattr(obj, 'help_box')) or (obj.help_box is None) or (not obj.help_box.winfo_exists()):
        obj.help_box = HelpBox(text=text, image=image, font=font)  # create window if its None or destroyed
        obj.help_box.after(100, obj.help_box.focus)  # Workaround for bug where main window takes focus
    else:
        obj.help_box.text(text, image)  # change the text and image


class HelpWindowMain :
    # class that can create a HelpBox (a windows of help)
    def __init__(self, font=None):
        self.help_box = None
        self.font = font

    def create_help_box(self, text, image=None, font=None):
        if (not hasattr(self, 'help_box')) or (self.help_box is None) or (not self.help_box.winfo_exists()):
            if font is None and self.font is not None :
                font = self.font

            self.help_box = HelpBox(text=text, image=image, font=font)  # create window if its None or destroyed
            self.help_box.after(100, self.help_box.focus)  # Workaround for bug where main window takes focus
        else:
            self.help_box.text(text, image)  # change the text



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2020 Michael J. Hayford
"""Aberration calculations for (fld, wvl, foc), including focus and image shift

    This module refactors some existing ray trace and aberration calculations
    in other modules to be expressed for a single field point and wavelength.
    The ability to apply focus and image shifts to an already acquired data set
    is provided for use interactively and in other performance critical areas.

    The following classes are implemented in this module:

        - :class:`~.Ray`: trace a single ray
        - :class:`~.RayFan`: trace a fan of rays in either the x or y meridian
        - :class:`~.RayList`: trace a list of rays from an object point
        - :class:`~.RayGrid`: trace a rectilinear grid of rays

    All but the `Ray` class are supported by a group of functions to trace the
    rays, accumulate the data (trace_*), and refocus (focus_*) the data. A
    all-in-one function (eval_*) to trace and apply focus is supplied also.
    These are used in the update_data methods of the classes to generate the
    ray data.

    This module also has functions to calculate chief ray and reference sphere
    information as well as functions for calculating the monochromatic PSF of
    the model.

.. Created on Sat Feb 22 22:01:56 2020

.. codeauthor: Michael J. Hayford
"""

# fork of rayoptics 0.9.3 https://github.com/mjhoptics/ray-optics/tree/master
# modify functions to compute wavefront error maps on the reference sphere of another field/wavelength

import numpy as np
from numpy.fft import fftshift, fft2

from scipy.interpolate import interp1d

import rayoptics.optical.model_constants as mc

from rayoptics.raytr import sampler
from rayoptics.raytr import trace
from rayoptics.raytr import traceerror as terr
from rayoptics.raytr import waveabr


# --- Single ray
class Ray():
    """A ray at the given field and wavelength.

    Attributes:
        opt_model: :class:`~.OpticalModel` instance
        p: relative 2d pupil coordinates
        f: index into :class:`~.FieldSpec` or a :class:`~.Field` instance
        wl: wavelength (nm) to trace the ray, or central wavelength if None
        foc: focus shift to apply to the results
        image_pt_2d: base image point. if None, the chief ray is used
        image_delta: image offset to apply to image_pt_2d
        srf_save:

            'single': save the ray data for surface srf_indx
            'all': save all of the surface by surface ray data

        srf_indx: for single surface retention, the surface index to save
    """

    def __init__(self, opt_model, p, f=0, wl=None, foc=None, image_pt_2d=None,
                 image_delta=None, srf_indx=-1, srf_save='single', 
                 output_filter=None, rayerr_filter=None, color=None, 
                 clip_rays=False):
        self.opt_model = opt_model
        osp = opt_model.optical_spec
        self.pupil = p
        self.fld = osp['fov'].fields[f] if isinstance(f, int) else f
        self.wvl = osp['wvls'].central_wvl if wl is None else wl

        self.foc = osp['focus'].focus_shift if foc is None else foc
        self.image_pt_2d = image_pt_2d
        self.image_delta = image_delta

        self.output_filter = output_filter
        self.rayerr_filter = rayerr_filter
        self.clip_rays = clip_rays

        self.color = color

        self.srf_save = srf_save
        self.srf_indx = srf_indx

        self.update_data()

    def update_data(self, **kwargs):
        """Trace the ray and calculate transverse aberrations. """
        ref_sphere, cr_pkg = trace.setup_pupil_coords(
            self.opt_model, self.fld, self.wvl, self.foc, 
            image_pt=self.image_pt_2d, image_delta=self.image_delta
            )
        build = kwargs.pop('build', 'rebuild')
        if build == 'rebuild':
            ray_result = trace.trace_safe(
                self.opt_model, self.pupil, self.fld, self.wvl, 
                self.output_filter, self.rayerr_filter, 
                use_named_tuples=True, check_apertures=self.clip_rays, 
                **kwargs)
            ray_pkg, ray_err = ray_result
            self.ray_seg = ray_pkg.ray[self.srf_indx]

            if self.srf_save == 'all':
                self.ray_pkg = ray_pkg


        ray_seg = self.ray_seg
        dist = self.foc / ray_seg[mc.d][2]
        defocused_pt = ray_seg[mc.p] + dist*ray_seg[mc.d]
        reference_image_pt = ref_sphere[0]
        self.t_abr = defocused_pt[:2] - reference_image_pt[:2]

        return self


# --- Fan of rays
class RayFan():
    """A fan of rays across the pupil at the given field and wavelength.

    Attributes:
        opt_model: :class:`~.OpticalModel` instance
        f: index into :class:`~.FieldSpec` or a :class:`~.Field` instance
        wl: wavelength (nm) to trace the fan, or central wavelength if None
        foc: focus shift to apply to the results
        image_pt_2d: base image point. if None, the chief ray is used
        image_delta: image offset to apply to image_pt_2d
        num_rays: number of samples along the fan
        xyfan: 'x' or 'y', specifies the axis the fan is sampled on
    """

    def __init__(self, opt_model, f=0, wl=None, foc=None, image_pt_2d=None,
                 image_delta=None, num_rays=21, xyfan='y', output_filter=None,
                 rayerr_filter=None, color=None, clip_rays=False, 
                 **kwargs):
        self.opt_model = opt_model
        osp = opt_model.optical_spec
        self.fld = osp.field_of_view.fields[f] if isinstance(f, int) else f
        self.wvl = osp.spectral_region.central_wvl if wl is None else wl

        self.foc = osp.defocus.focus_shift if foc is None else foc
        self.image_pt_2d = image_pt_2d
        self.image_delta = image_delta

        self.num_rays = num_rays

        if xyfan == 'x':
            self.xyfan = 0
        elif xyfan == 'y':
            self.xyfan = 1
        else:
            self.xyfan = int(xyfan)

        self.color = color

        self.rt_kwargs = kwargs
        self.rt_kwargs['output_filter'] = output_filter
        self.rt_kwargs['rayerr_filter'] = rayerr_filter
        self.rt_kwargs['check_apertures'] = clip_rays

        self.update_data()

    def __json_encode__(self):
        attrs = dict(vars(self))
        del attrs['opt_model']
        del attrs['fan_pkg']
        return attrs

    def update_data(self, **kwargs):
        """Set the fan attribute to a list of (pupil coords), dx, dy, opd."""
        build = kwargs.get('build', 'rebuild')
        if build == 'rebuild':
            self.fan_pkg = trace_fan(
                self.opt_model, self.fld, self.wvl, self.foc, self.xyfan,
                image_pt_2d=self.image_pt_2d, image_delta=self.image_delta, 
                num_rays=self.num_rays,
                **self.rt_kwargs)

        self.fan = focus_fan(self.opt_model, self.fan_pkg,
                             self.fld, self.wvl, self.foc,
                             image_pt_2d=self.image_pt_2d,
                             image_delta=self.image_delta,
                             **self.rt_kwargs)
        return self


def select_plot_data(fan, xyfan, data_type):
    """Given a fan of data, select the sample points and the resulting data."""
    f_x = []
    f_y = []
    for p, val in fan:
        f_x.append(p[xyfan])
        f_y.append(val[data_type])
    f_x = np.array(f_x)
    f_y = np.array(f_y)
    return f_x, f_y


def smooth_plot_data(f_x, f_y, num_points=100):
    """Interpolate fan data points and return a smoothed version."""
    interpolator = interp1d(f_x, f_y,
                            kind='cubic', assume_sorted=True)
    x_sample = np.linspace(f_x.min(), f_x.max(), num_points)
    y_fit = interpolator(x_sample)

    return x_sample, y_fit


def trace_ray_fan(opt_model, fan_rng, fld, wvl, foc,
                  output_filter=None, rayerr_filter=None, **kwargs):
    """Trace a fan of rays, according to fan_rng. """
    start = np.array(fan_rng[0])
    stop = fan_rng[1]
    num = fan_rng[2]
    step = (stop - start)/(num - 1)
    fan = []
    for r in range(num):
        pupil = np.array(start)
        ray_result = trace.trace_safe(opt_model, pupil, fld, wvl, 
                                      output_filter, rayerr_filter, 
                                      use_named_tuples=True, **kwargs)
        ray_pkg, ray_err = ray_result

        if ray_pkg is not None:
            fan.append([pupil[0], pupil[1], ray_pkg])
        start += step
    return fan


def eval_fan(opt_model, fld, wvl, foc, xy,
             image_pt_2d=None, image_delta=None, num_rays=21,
             output_filter=None, rayerr_filter=None, **kwargs):
    """Trace a fan of rays and evaluate dx, dy, & OPD across the fan."""
    fod = opt_model['analysis_results']['parax_data'].fod
    ref_sphere, cr_pkg = trace.setup_pupil_coords(opt_model, fld, wvl, foc, 
                                                  image_pt=image_pt_2d,
                                                  image_delta=image_delta)
    fld.chief_ray = cr_pkg
    fld.ref_sphere = ref_sphere

    fan_start = np.array([0., 0.])
    fan_stop = np.array([0., 0.])
    fan_start[xy] = -1.0
    fan_stop[xy] = 1.0
    fan_def = [fan_start, fan_stop, num_rays]

    fan = trace_ray_fan(opt_model, fan_def, fld, wvl, foc,
                        output_filter=output_filter,
                        rayerr_filter=rayerr_filter, **kwargs)

    central_wvl = opt_model.optical_spec.spectral_region.central_wvl
    convert_to_opd = 1/opt_model.nm_to_sys_units(central_wvl)

    def rfc(fi):
        pupil_x, pupil_y, ray_pkg = fi
        if ray_pkg is not None:
            image_pt = ref_sphere[0]
            ray = ray_pkg[mc.ray]
            dist = foc / ray[-1][mc.d][2]
            defocused_pt = ray[-1][mc.p] + dist*ray[-1][mc.d]
            t_abr = defocused_pt - image_pt

            opdelta = waveabr.wave_abr_full_calc(fod, fld, wvl, foc, ray_pkg,
                                                 cr_pkg, ref_sphere)
            opd = convert_to_opd*opdelta
            return (pupil_x, pupil_y), (t_abr[0], t_abr[1], opd)
        else:
            return pupil_x, pupil_y, np.nan
    fan_data = [rfc(i) for i in fan]

    return fan_data


def trace_fan(opt_model, fld, wvl, foc, xy,
              image_pt_2d=None, image_delta=None, num_rays=21,
              output_filter=None, rayerr_filter=None, **kwargs):
    """Trace a fan of rays and precalculate data for rapid refocus later."""
    fod = opt_model['analysis_results']['parax_data'].fod
    ref_sphere, cr_pkg = trace.setup_pupil_coords(opt_model, fld, wvl, foc, 
                                                  image_pt=image_pt_2d,
                                                  image_delta=image_delta)
    fld.chief_ray = cr_pkg
    fld.ref_sphere = ref_sphere

    """ xy determines whether x (=0) or y (=1) fan """
    fan_start = np.array([0., 0.])
    fan_stop = np.array([0., 0.])
    fan_start[xy] = -1.0
    fan_stop[xy] = 1.0
    fan_def = [fan_start, fan_stop, num_rays]

    fan = trace_ray_fan(opt_model, fan_def, fld, wvl, foc,
                        output_filter=output_filter,
                        rayerr_filter=rayerr_filter, **kwargs)

    def wpc(fi):
        pupil_x, pupil_y, ray_pkg = fi
        if ray_pkg is not None and not isinstance(ray_pkg, terr.TraceError):
            pre_opd_pkg = waveabr.wave_abr_pre_calc(fod, fld, wvl, foc,
                                                    ray_pkg, cr_pkg, 
                                                    ref_sphere)
            return pre_opd_pkg
        else:
            return None

    upd_fan = [wpc(i) for i in fan]

    return fan, upd_fan


def focus_fan(opt_model, fan_pkg, fld, wvl, foc, 
              image_pt_2d=None, image_delta=None, **kwargs):
    """Refocus the fan of rays and return the tranverse abr. and OPD."""
    fod = opt_model['analysis_results']['parax_data'].fod
    fan, upd_fan = fan_pkg
    ref_sphere, cr_pkg = trace.setup_pupil_coords(opt_model, fld, wvl, foc, 
                                                  image_pt=image_pt_2d,
                                                  image_delta=image_delta)
    central_wvl = opt_model.optical_spec.spectral_region.central_wvl
    convert_to_opd = 1/opt_model.nm_to_sys_units(central_wvl)

    def rfc(fi, fiu):
        pupil_x, pupil_y, ray_pkg = fi
        if ray_pkg is not None and not isinstance(ray_pkg, terr.TraceError):
            image_pt = ref_sphere[0]
            ray = ray_pkg[mc.ray]
            dist = foc / ray[-1][mc.d][2]
            defocused_pt = ray[-1][mc.p] + dist*ray[-1][mc.d]
            t_abr = defocused_pt - image_pt

            opdelta = waveabr.wave_abr_calc(fod, fld, wvl, foc,
                                            ray_pkg, cr_pkg, fiu, ref_sphere)
            opd = convert_to_opd*opdelta
            return (pupil_x, pupil_y), (t_abr[0], t_abr[1], opd)
        else:
            return pupil_x, pupil_y, np.nan
    fan_data = [rfc(fi, fiu) for fi, fiu in zip(fan, upd_fan)]
    return fan_data


# --- List of rays
class RayList():
    """Container class for a list of rays produced from a list or generator

    Attributes:
        opt_model: :class:`~.OpticalModel` instance
        pupil_gen: (fct, args, kwargs), where:

            - fct: a function returning a generator returning a 2d coordinate
            - args: passed to fct
            - kwargs: passed to fct

        pupil_coords: list of 2d coordinates. If None, filled in by calling
                      pupil_gen.
        num_rays: number of samples side of grid. Used only if pupil_coords and
                  pupil_gen are None.
        f: index into :class:`~.FieldSpec` or a :class:`~.Field` instance
        wl: wavelength (nm) to trace the fan, or central wavelength if None
        foc: focus shift to apply to the results
        image_pt_2d: base image point. if None, the chief ray is used
        image_delta: image offset to apply to image_pt_2d
        apply_vignetting: whether to apply vignetting factors to pupil coords
    """

    def __init__(self, opt_model,
                 pupil_gen=None, pupil_coords=None, num_rays=21,
                 f=0, wl=None, foc=None, image_pt_2d=None, image_delta=None, 
                 output_filter=None, rayerr_filter=None, clip_rays=False, 
                 apply_vignetting=True, **kwargs):
        self.opt_model = opt_model
        osp = opt_model.optical_spec
        if pupil_coords is not None and pupil_gen is None:
            self.pupil_coords = pupil_coords
            self.pupil_gen = None
        else:
            if pupil_gen is not None:
                self.pupil_gen = pupil_gen
            else:
                grid_start = np.array([-1., -1.])
                grid_stop = np.array([1., 1.])
                grid_def = [grid_start, grid_stop, num_rays]
                self.pupil_gen = (sampler.csd_grid_ray_generator,
                                  (grid_def,), {})
            fct, args, kwa = self.pupil_gen
            self.pupil_coords = fct(*args, **kwa)

        self.fld = osp.field_of_view.fields[f] if isinstance(f, int) else f
        self.wvl = osp.spectral_region.central_wvl if wl is None else wl

        self.foc = osp.defocus.focus_shift if foc is None else foc
        self.image_pt_2d = image_pt_2d
        self.image_delta = image_delta

        self.rt_kwargs = kwargs
        self.rt_kwargs['apply_vignetting'] = apply_vignetting
        self.rt_kwargs['output_filter'] = output_filter
        self.rt_kwargs['rayerr_filter'] = rayerr_filter
        self.rt_kwargs['check_apertures'] = clip_rays

        self.update_data()

    def __json_encode__(self):
        attrs = dict(vars(self))
        del attrs['opt_model']
        del attrs['pupil_gen']
        del attrs['pupil_coords']
        del attrs['ray_list']
        return attrs

    def update_data(self, **kwargs):
        build = kwargs.get('build', 'rebuild')
        if build == 'rebuild':
            if self.pupil_gen:
                fct, args, kwa = self.pupil_gen
                self.pupil_coords = fct(*args, **kwa)

            self.ray_list = trace_pupil_coords(
                self.opt_model, self.pupil_coords,
                self.fld, self.wvl, self.foc,
                image_pt_2d=self.image_pt_2d, 
                image_delta=self.image_delta, 
                **self.rt_kwargs)

        ray_list_data = focus_pupil_coords(
            self.opt_model, self.ray_list,
            self.fld, self.wvl, self.foc,
            image_pt_2d=self.image_pt_2d,
            image_delta=self.image_delta,
            **self.rt_kwargs)

        self.ray_abr = np.rollaxis(ray_list_data, 1)

        return self


def trace_ray_list(opt_model, pupil_coords, fld, wvl, foc,
                   append_if_none=False,
                   output_filter=None, rayerr_filter=None,
                   **kwargs):
    """Trace a list of rays at fld and wvl and return ray_pkgs in a list."""

    ray_list = []
    for pupil in pupil_coords:
        ray_result = trace.trace_safe(opt_model, pupil, fld, wvl, 
                                      output_filter, rayerr_filter, 
                                      **kwargs)
        ray_pkg, ray_err = ray_result
        if ray_pkg is not None:
            ray_list.append([pupil[0], pupil[1], ray_pkg])
        else:  # ray outside pupil or failed
            if append_if_none:
                ray_list.append([pupil[0], pupil[1], None])

    return ray_list


def trace_list_of_rays(opt_model, rays,
                       output_filter=None, rayerr_filter=None,
                       **kwargs):
    """Trace a list of rays (pt, dir, wvl) and return ray_pkgs in a list.

    Args:
        opt_model: :class:`~.OpticalModel` instance
        rays: list of (pt0, dir0, wvl)

            - pt0: starting point in coords of first interface
            - dir0: starting direction cosines in coords of first interface
            - wvl: wavelength in nm

        output_filter: None, "last", or a callable. See below
        **kwargs: kwyword args passed to the trace function

    The output_filter keyword argument controls what ray data is returned to
    the caller.

        - if None, returns the entire traced ray
        - if "last", returns the ray data from the last interface
        - if a callable, it must take a ray_pkg as an argument and return the
          desired data or None

    Returns:
        A list with an entry for each ray in rays
    """
    ray_list = []
    for ray in rays:
        pt0, dir0, wvl = ray
        try:
            ray_pkg = trace.trace(opt_model.seq_model, pt0, dir0, wvl, 
                                  **kwargs)
        except terr.TraceError as rayerr:
            if rayerr_filter is None:
                pass
            elif rayerr_filter == 'full':
                ray_list.append((ray, rayerr))
            elif rayerr_filter == 'summary':
                rayerr.ray_pkg = None
                ray_list.append((ray, rayerr))
            else:
                pass
        else:
            if output_filter is None:
                ray_list.append(ray_pkg)
            elif output_filter == 'last':
                ray, op_delta, wvl = ray_pkg
                ray_list.append((ray[-1], op_delta, wvl))
            else:
                ray_list.append(output_filter(ray_pkg))

    return ray_list


def eval_pupil_coords(opt_model, fld, wvl, foc, image_pt_2d=None, 
                      image_delta=None, num_rays=21, **kwargs):
    """Trace a list of rays and return the transverse abr."""
    ref_sphere, cr_pkg = trace.setup_pupil_coords(opt_model, fld, wvl, foc, 
                                                  image_pt=image_pt_2d,
                                                  image_delta=image_delta)
    fld.chief_ray = cr_pkg
    fld.ref_sphere = ref_sphere

    grid_start = np.array([-1., -1.])
    grid_stop = np.array([1., 1.])
    grid_def = [grid_start, grid_stop, num_rays]

    kwargs['check_apertures'] = kwargs.get('check_apertures', True)
    ray_list = trace_ray_list(opt_model, sampler.grid_ray_generator(grid_def),
                              fld, wvl, foc, **kwargs)

    def rfc(ri):
        pupil_x, pupil_y, ray_pkg = ri
        if ray_pkg is not None:
            image_pt = ref_sphere[0]
            ray = ray_pkg[mc.ray]
            dist = foc / ray[-1][mc.d][2]
            defocused_pt = ray[-1][mc.p] + dist*ray[-1][mc.d]
            t_abr = defocused_pt - image_pt
            return t_abr[0], t_abr[1]
        else:
            return np.nan
    ray_list_data = [rfc(ri) for ri in ray_list]
    return np.array(ray_list_data)


def trace_pupil_coords(opt_model, pupil_coords, fld, wvl, foc,
                       ref_fld=None, ref_wvl=None,
                       image_pt_2d=None, image_delta=None, **kwargs):
    """Trace a list of rays and return data needed for rapid refocus."""

    ref_sphere, cr_pkg = trace.setup_pupil_coords(opt_model, fld, wvl, foc,
                                                  image_pt=image_pt_2d,
                                                  image_delta=image_delta)
    fld.chief_ray = cr_pkg
    fld.ref_sphere = ref_sphere

    kwargs['check_apertures'] = kwargs.get('check_apertures', True)
    ray_list = trace_ray_list(opt_model, pupil_coords,
                              fld, wvl, foc, **kwargs)

    return ray_list


def focus_pupil_coords(opt_model, ray_list, fld, wvl, foc, 
                       image_pt_2d=None, image_delta=None, **kwargs):
    """Given pre-traced rays and a ref. sphere, return the transverse abr."""
    ref_sphere, cr_pkg = trace.setup_pupil_coords(opt_model, fld, wvl, foc, 
                                                  image_pt=image_pt_2d,
                                                  image_delta=image_delta)

    def rfc(ri):
        pupil_x, pupil_y, ray_pkg = ri
        if ray_pkg is not None:
            image_pt = ref_sphere[0]
            ray = ray_pkg[mc.ray]
            dist = foc / ray[-1][mc.d][2]
            defocused_pt = ray[-1][mc.p] + dist*ray[-1][mc.d]
            t_abr = defocused_pt - image_pt
            return t_abr[0], t_abr[1]
        else:
            return np.nan
    ray_list_data = [rfc(ri) for ri in ray_list]
    return np.array(ray_list_data)


# --- Square grid of rays
class RayGrid():
    """Container for a square grid of rays.

    Attributes:
        opt_model: :class:`~.OpticalModel` instance
        f: index into :class:`~.FieldSpec` or a :class:`~.Field` instance
        wl: wavelength (nm) to trace the fan, or central wavelength if None
        foc: focus shift to apply to the results
        image_pt_2d: base image point. if None, the chief ray is used
        image_delta: image offset to apply to image_pt_2d
        num_rays: number of samples along the side of the grid
    """

    def __init__(self, opt_model, f=0, wl=None, foc=None, image_pt_2d=None,
                 image_delta=None, output_filter=None, rayerr_filter=None, 
                 num_rays=21, clip_rays=True, value_if_none=np.nan,
                 **kwargs):
        self.opt_model = opt_model
        osp = opt_model.optical_spec
        self.fld = osp.field_of_view.fields[f] if isinstance(f, int) else f
        self.wvl = osp.spectral_region.central_wvl if wl is None else wl

        self.foc = osp.defocus.focus_shift if foc is None else foc
        self.image_pt_2d = image_pt_2d
        self.image_delta = image_delta

        self.num_rays = num_rays
        self.value_if_none = value_if_none

        self.rt_kwargs = kwargs
        self.rt_kwargs['output_filter'] = output_filter
        self.rt_kwargs['rayerr_filter'] = rayerr_filter
        self.rt_kwargs['check_apertures'] = clip_rays

        self.update_data()

    def __json_encode__(self):
        attrs = dict(vars(self))
        del attrs['opt_model']
        del attrs['grid_pkg']
        return attrs

    def update_data(self, **kwargs):
        build = kwargs.get('build', 'rebuild')
        if build == 'rebuild':
            self.grid_pkg = trace_wavefront(
                self.opt_model, self.fld, self.wvl, self.foc,
                image_pt_2d=self.image_pt_2d, image_delta=self.image_delta, 
                num_rays=self.num_rays,
                **self.rt_kwargs)

        opd = focus_wavefront(self.opt_model, self.grid_pkg,
                              self.fld, self.wvl, self.foc,
                              image_pt_2d=self.image_pt_2d,
                              image_delta=self.image_delta, 
                              value_if_none=self.value_if_none,
                              **self.rt_kwargs)

        self.grid = np.rollaxis(opd, 2)

        return self


def trace_ray_grid(opt_model, grid_rng, fld, wvl, foc, append_if_none=True,
                   output_filter=None, rayerr_filter=None, **kwargs):
    """Trace a grid of rays at fld and wvl and return ray_pkgs in 2d list."""
    start = np.array(grid_rng[0])
    stop = grid_rng[1]
    num = grid_rng[2]
    step = np.array((stop - start)/(num - 1))
    grid = []
    kwargs['apply_vignetting'] = kwargs.get('apply_vignetting', False)
    for i in range(num):
        grid_row = []

        for j in range(num):
            pupil = np.array(start)
            ray_result = trace.trace_safe(opt_model, pupil, fld, wvl, 
                                          output_filter, rayerr_filter, 
                                          **kwargs)
            ray_pkg, ray_err = ray_result
            if ray_pkg is not None:
                    grid_row.append([pupil[0], pupil[1], ray_pkg])
            else:  # ray outside pupil or failed
                if append_if_none:
                    grid_row.append([pupil[0], pupil[1], None])

            start[1] += step[1]

        grid.append(grid_row)
        start[0] += step[0]
        start[1] = grid_rng[0][1]

    return grid


def eval_wavefront(opt_model, fld, wvl, foc,
                   ref_fld=None, ref_wvl=None,
                   image_pt_2d=None,
                   image_delta=None, num_rays=21, value_if_none=np.nan,
                   return_piston=False, remove_piston=False,
                   **kwargs):
    """Trace a grid of rays and evaluate the OPD across the wavefront."""
    fod = opt_model['analysis_results']['parax_data'].fod

    if ref_wvl is None:
        ref_wvl = wvl
    if ref_fld is None:
        ref_fld = fld

    ref_sphere, cr_pkg = trace.setup_pupil_coords(opt_model, ref_fld, ref_wvl, foc,
                                                  image_pt=image_pt_2d,
                                                  image_delta=image_delta)

    fld.chief_ray = cr_pkg
    fld.ref_sphere = ref_sphere

    vig_bbox = fld.vignetting_bbox(opt_model['osp']['pupil'])
    vig_grid_def = [vig_bbox[0], vig_bbox[1], num_rays]

    kwargs['check_apertures'] = kwargs.get('check_apertures', True)
    grid = trace_ray_grid(opt_model, vig_grid_def, 
                          fld, wvl, foc, **kwargs)

    central_wvl = opt_model.optical_spec.spectral_region.central_wvl
    # convert_to_opd = 1/opt_model.nm_to_sys_units(central_wvl)
    convert_to_opd = 1 / opt_model.nm_to_sys_units(wvl)

    # determine piston to have the chief ray with a 0 offset value (piston as center of pupil has no offset)
    ray00_pkg, ray00_err = trace.trace_safe(opt_model, [0, 0], fld, wvl, None, None)
    piston_d = waveabr.wave_abr_full_calc(fod, fld, wvl, foc, ray00_pkg, cr_pkg, ref_sphere)
    piston = convert_to_opd * piston_d


    def rfc(gij):
        pupil_x, pupil_y, ray_pkg = gij
        if ray_pkg is not None:
            opdelta = waveabr.wave_abr_full_calc(fod, fld, wvl, foc, ray_pkg,
                                                 cr_pkg, ref_sphere)
            opd = convert_to_opd*opdelta

            if remove_piston :
                opd = opd - piston

            return pupil_x, pupil_y, opd
        else:
            return pupil_x, pupil_y, value_if_none
    opd_grid = [[rfc(j) for j in i] for i in grid]

    if return_piston :
        return np.array(opd_grid), piston
    else :
        return np.array(opd_grid)


def trace_wavefront(opt_model, fld, wvl, foc,
                    ref_fld=None, ref_wvl=None,
                    image_pt_2d=None, image_delta=None, num_rays=21,
                    return_piston=False,
                    **kwargs):
    """
    Trace a grid of rays and pre-calculate data needed for rapid refocus.
    If return_piston, pre-calculate the piston for the ray passing by pupil center
    """
    fod = opt_model['analysis_results']['parax_data'].fod

    if ref_wvl is None:
        ref_wvl = wvl
    if ref_fld is None:
        ref_fld = fld

    ref_sphere, cr_pkg = trace.setup_pupil_coords(opt_model, ref_fld, ref_wvl, foc,
                                                  image_pt=image_pt_2d,
                                                  image_delta=image_delta)

    fld.chief_ray = cr_pkg
    fld.ref_sphere = ref_sphere

    vig_bbox = fld.vignetting_bbox(opt_model['osp']['pupil'])
    vig_grid_def = [vig_bbox[0], vig_bbox[1], num_rays]

    kwargs['check_apertures'] = kwargs.get('check_apertures', True)
    grid = trace_ray_grid(opt_model, vig_grid_def, 
                          fld, wvl, foc, **kwargs)

    def wpc(gij):
        pupil_x, pupil_y, ray_pkg = gij
        if ray_pkg is not None:
            pre_opd_pkg = waveabr.wave_abr_pre_calc(fod, fld, wvl, foc,
                                                    ray_pkg, cr_pkg, 
                                                    ref_sphere)
            return pre_opd_pkg
        else:
            return None
    upd_grid = [[wpc(j) for j in i] for i in grid]

    if return_piston :
        # determine piston to have the chief ray with a 0 offset value (piston as center of pupil has no offset)
        ray00_pkg, ray00_err = trace.trace_safe(opt_model, [0, 0], fld, wvl, None, None)
        pre_piston = waveabr.wave_abr_pre_calc(fod, fld, wvl, foc, ray00_pkg, cr_pkg, ref_sphere)
        return grid, upd_grid, ray00_pkg, pre_piston
    else:
        return grid, upd_grid


def focus_wavefront(opt_model, grid_pkg, fld, wvl, foc,
                    ref_fld=None, ref_wvl=None,
                    image_pt_2d=None,
                    image_delta=None, value_if_none=np.nan,
                    piston_pkg=None, return_piston=False, **kwargs):
    """
    Given pre-traced rays and a ref. sphere, return the ray's OPD.
    piston_pkg is not None, remove a piston value corresponding to piston_pkg
    piston_pkg can be obtained by trace_wavefront, and is [ray00_pkg, pre_piston]
    """
    fod = opt_model['analysis_results']['parax_data'].fod
    grid, upd_grid = grid_pkg

    if ref_wvl is None:
        ref_wvl = wvl
    if ref_fld is None:
        ref_fld = fld

    ref_sphere, cr_pkg = trace.setup_pupil_coords(opt_model, ref_fld, ref_wvl, foc,
                                                  image_pt=image_pt_2d,
                                                  image_delta=image_delta)

    convert_to_opd = 1 / opt_model.nm_to_sys_units(wvl)

    if (piston_pkg is not None) and (None not in piston_pkg):
        ray00_pkg, pre_piston = piston_pkg
        # determine piston to have the chief ray with a 0 offset value (piston as center of pupil has no offset)
        piston_d = waveabr.wave_abr_calc(fod, fld, wvl, foc, ray00_pkg, cr_pkg, pre_piston, ref_sphere)
        piston = convert_to_opd * piston_d
    else :
        piston = None

    def rfc(gij, uij):
        pupil_x, pupil_y, ray_pkg = gij
        if ray_pkg is not None:
            opdelta = waveabr.wave_abr_calc(fod, fld, wvl, foc,
                                            ray_pkg, cr_pkg, uij, ref_sphere)
            opd = convert_to_opd*opdelta

            if piston is not None:
                opd = opd - piston

            return pupil_x, pupil_y, opd
        else:
            return pupil_x, pupil_y, value_if_none
    refocused_grid = [[rfc(jg, ju) for jg, ju in zip(ig, iu)]
                      for ig, iu in zip(grid, upd_grid)]

    if return_piston :
        return np.array(refocused_grid), piston
    else:
        return np.array(refocused_grid)


# --- PSF calculation
def psf_sampling(n=None, n_pupil=None, n_airy=None):
    """Given 2 of 3 parameters, calculate the third.

    Args:
        n: The total width of the sampling grid
        n_pupil: The sampling across the pupil
        n_airy: The sampling across the central peak of the Airy disk

    Returns: (n, n_pupil, n_airy)
    """
    npa = n, n_pupil, n_airy
    i = npa.index(None)
    if i == 0:
        n = round((n_pupil * n_airy)/2.44)
    elif i == 1:
        n_pupil = round(2.44*n/n_airy)
    elif i == 2:
        n_airy = round(2.44*n/n_pupil)
    return n, n_pupil, n_airy


def calc_psf_scaling(pupil_grid, ndim, maxdim):
    """Calculate the input and output grid spacings.

    Args:
        pupil_grid: A RayGrid instance
        ndim: The sampling across the wavefront
        maxdim: The total width of the sampling grid

    Returns:
        delta_x: The linear grid spacing on the entrance pupil
        delta_xp: The linear grid spacing on the image plane
    """
    opt_model = pupil_grid.opt_model
    fod = opt_model['analysis_results']['parax_data'].fod
    wl = opt_model.nm_to_sys_units(pupil_grid.wvl)

    fill_factor = ndim/maxdim
    max_D = 2 * fod.enp_radius / fill_factor
    delta_x = max_D / maxdim
    C = wl/fod.exp_radius

    delta_theta = (fill_factor * C) / 2
    ref_sphere_radius = pupil_grid.fld.ref_sphere[2]
    delta_xp = delta_theta * ref_sphere_radius

    return delta_x, delta_xp


def calc_psf(wavefront, ndim, maxdim):
    """Calculate the point spread function of wavefront W.

    Args:
        wavefront: ndim x ndim Numpy array of wavefront errors. No data
                   condition is indicated by nan
        ndim: The sampling across the wavefront
        maxdim: The total width of the sampling grid

    Returns: AP, the PSF of the input wavefront
    """
    maxdim_by_2 = maxdim//2
    W = np.zeros([maxdim, maxdim])
    nd2 = ndim//2
    W[maxdim_by_2-(nd2-1):maxdim_by_2+(nd2+1),
      maxdim_by_2-(nd2-1):maxdim_by_2+(nd2+1)] = np.nan_to_num(wavefront)

    phase = np.exp(1j*2*np.pi*W)

    for i in range(len(phase)):
        for j in range(len(phase)):
            if phase[i][j] == 1:
                phase[i][j] = 0

    AP = abs(fftshift(fft2(fftshift(phase))))**2
    AP_max = np.nanmax(AP)
    AP = AP/AP_max
    return AP


def update_psf_data(pupil_grid, build='rebuild'):
    pupil_grid.update_data(build=build)
    ndim = pupil_grid.num_rays
    maxdim = pupil_grid.maxdim
    AP = calc_psf(pupil_grid.grid[2], ndim, maxdim)
    return AP
